@@ -75,9 +75,12 @@ SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 #include "compiler.h"
 
 #if HAS_DEBUG_FULL
-#define DBG(x) ErrorF x
+void LogF(const char *f, ...);
+#define DBG(x) LogF x
+#define ERR(x) ErrorF x
 #else
 #define DBG(x)
+#define ERR(x)
 #endif
 
 #define DEBUG_NO_BLT 0
@@ -142,10 +145,11 @@ struct sna_pixmap {
 
 #define SOURCE_BIAS 4
 	uint16_t source_count;
-	uint8_t pinned :3;
+	uint8_t pinned :4;
 #define PIN_SCANOUT 0x1
-#define PIN_DRI 0x2
-#define PIN_PRIME 0x4
+#define PIN_DRI2 0x2
+#define PIN_DRI3 0x4
+#define PIN_PRIME 0x8
 	uint8_t create :4;
 	uint8_t mapped :2;
 #define MAPPED_NONE 0
@@ -204,7 +208,7 @@ struct sna_gc {
 	long changes;
 	long serial;
 
-	GCFuncs *old_funcs;
+	const GCFuncs *old_funcs;
 	void *priv;
 };
 
@@ -315,6 +319,7 @@ struct sna {
 		struct gen5_render_state gen5;
 		struct gen6_render_state gen6;
 		struct gen7_render_state gen7;
+		struct gen8_render_state gen8;
 	} render_state;
 
 	bool dri_available;
@@ -328,7 +333,7 @@ struct sna {
 
 #if HAVE_UDEV
 	struct udev_monitor *uevent_monitor;
-	InputHandlerProc uevent_handler;
+	pointer uevent_handler;
 #endif
 
 	struct {
@@ -343,6 +348,7 @@ struct sna {
 #if DEBUG_MEMORY
 	struct {
 		int pixmap_allocs;
+		int pixmap_cached;
 		int cpu_bo_allocs;
 		size_t shadow_pixels_bytes;
 		size_t cpu_bo_bytes;
@@ -501,20 +507,23 @@ PixmapPtr sna_pixmap_create_unattached(ScreenPtr screen,
 				       int width, int height, int depth);
 void sna_pixmap_destroy(PixmapPtr pixmap);
 
-#define assert_pixmap_map(pixmap, priv) \
-	assert(priv->mapped == false || pixmap->devPrivate.ptr == (priv->mapped == MAPPED_CPU ? MAP(priv->gpu_bo->map__cpu) : MAP(priv->gpu_bo->map__gtt)));
+#define assert_pixmap_map(pixmap, priv)  do { \
+	assert(priv->mapped != MAPPED_NONE || pixmap->devPrivate.ptr == PTR(priv->ptr)); \
+	assert(priv->mapped == MAPPED_NONE || pixmap->devPrivate.ptr == (priv->mapped == MAPPED_CPU ? MAP(priv->gpu_bo->map__cpu) : MAP(priv->gpu_bo->map__gtt))); \
+} while (0)
 
 static inline void sna_pixmap_unmap(PixmapPtr pixmap, struct sna_pixmap *priv)
 {
-	if (priv->mapped == MAPPED_NONE)
+	if (priv->mapped == MAPPED_NONE) {
+		assert(pixmap->devPrivate.ptr == PTR(priv->ptr));
 		return;
+	}
 
 	DBG(("%s: pixmap=%ld dropping %s mapping\n",
 	     __FUNCTION__, pixmap->drawable.serialNumber,
 	     priv->mapped == MAPPED_CPU ? "cpu" : "gtt"));
 
 	assert_pixmap_map(pixmap, priv);
-	assert(priv->stride && priv->stride);
 
 	pixmap->devPrivate.ptr = PTR(priv->ptr);
 	pixmap->devKind = priv->stride;
@@ -1020,4 +1029,7 @@ static inline void sigtrap_put(void)
 	--sigtrap;
 }
 
+#define RR_Rotate_All (RR_Rotate_0 | RR_Rotate_90 | RR_Rotate_180 | RR_Rotate_270)
+#define RR_Reflect_All (RR_Reflect_X | RR_Reflect_Y)
+
 #endif /* _SNA_H */
